import numpy as np
import matplotlib.pyplot as plt
#import matplotlib.pyplot as plt2
#import pandas as pd
from math import *
from pylab import mpl
mpl.rcParams['font.sans-serif'] = ['SimHei']
mpl.rcParams['axes.unicode_minus'] = False

class WSN(object):
    """ The network architecture with desired parameters """
    xm    = 100  # Length of the yard
    ym    = 100  # Width of the yard
    n     = 100  # total number of nodes
    sink  = None # Sink node
    nodes = None # All sensor nodes set
    # Energy model (all values in Joules)
    # Eelec = ETX = ERX
    ETX = 50 * (10 ** (-9))      # Energy for transferring of each bit
    ERX = 50 * (10 ** (-9))      # Energy for receiving of each bit
    # Transmit Amplifier types
    Efs = 10 * (10 ** (-12))     # Energy of free space model
    Emp = 0.0013 * (10 ** (-12)) # Energy of multi path model
    EDA = 5 * (10 ** (-9))       # Data aggregation energy
    f_r = 0.6                    # fusion_rate
    # Message
    CM = 20
    DM = 4000*2

    # computation of do
    do = np.sqrt(Efs / Emp) # 87.70580193070293

    #m_n = 3 # the number of malicious sensor nodes

    # Node State in Network
    n_dead           = 0 # The number of dead nodes
    flag_first_dead  = 0 # Flag tells that the first node died
    flag_all_dead    = 0 # Flag tells that all nodes died
    flag_net_stop    = 0 # Flag tells that network stop working:90% nodes died
    round_first_dead = 0 # The round when the first node died
    round_all_dead   = 0 # The round when all nodes died
    round_net_stop   = 0 # The round when the network stop working
    packets_data_d=0
    packets_data_s=0
    d_data=0
    def dist(x, y):
        distance = np.sqrt(np.power((x.xm - y.xm), 2) + np.power((x.ym - y.ym), 2))
        return distance

    def trans_energy(data, dis):
        if dis > WSN.do:
            energy = WSN.ETX * data + WSN.Emp * data * (dis ** 4)
        else: # min_dis <= do
            energy = WSN.ETX * data + WSN.Efs * data * (dis ** 2)
        return energy

    def node_state(r):
        nodes  = WSN.nodes
        n_dead = 0
        for node in nodes:
            if node.energy <= Node.energy_threshold:
                n_dead += 1
                WSN.packets_data_d+=1
                if WSN.flag_first_dead == 0 and n_dead == 1:
                    WSN.flag_first_dead  = 1
                    WSN.round_first_dead = r - Leach.r_empty
                    WSN.packets_data_d+=1;
                    WSN.d_data+=1;
        if WSN.flag_net_stop == 0 and n_dead >= (WSN.n * 0.9):
            WSN.flag_net_stop  = 1
            WSN.packets_data_s+=1;
            WSN.round_net_stop = r - Leach.r_empty
        if n_dead == WSN.n - 1:
            WSN.flag_all_dead  = 1
            WSN.round_all_dead = r - Leach.r_empty
        WSN.n_dead = n_dead

class Node(object):
    """ Sensor Node """
    energy_init      = 0.5 # initial energy of a node
    energy_threshold = 0.02

    def __init__(self):
        """ Create the node with default attributes """
        self.id      = None
        self.xm      = np.random.random() * WSN.xm
        self.ym      = np.random.random() * WSN.ym
        self.energy  = Node.energy_init
        self.type    = "N" #
        # G is the set of nodes that have not been cluster-heads in the last 1/p rounds.
        self.G       = 0 # the flag determines whether it's a CH or not
        self.head_id = None # The id of its CH

    def init_nodes():
        """ Initialize attributes of every node in order """
        nodes = []
        # Initial common node
        for i in range(WSN.n):
            node    = Node()
            node.id = i
            nodes.append(node)
        # Initial sink node
        sink    = Node()
        sink.id = -1
        sink.xm  = 0.5 * WSN.xm # x coordination of base station
        sink.ym  = 50 + WSN.ym # y coordination of base station
        # Add to WSN
        WSN.nodes = nodes
        WSN.sink  = sink

    def plot_wsn():
        nodes = WSN.nodes
        n = WSN.n
        #m_n = WSN.m_n
        # base station
        sink = WSN.sink
        plt.plot([sink.xm], [sink.ym], 'r^',label="Sink Node")
        #
        n_flag = True
        for i in range(n):
            if n_flag:
                plt.plot([nodes[i].xm], [nodes[i].ym], 'b+',label='Sensor Node')
                n_flag = False
            else:
                plt.plot([nodes[i].xm], [nodes[i].ym], 'b+')
        #
        '''
        m_flag = True
        for i in range(m_n):
            j = n + i
            if m_flag:
                plt.plot([nodes[j].xm], [nodes[j].ym], 'kd',label='aaaaa')
                m_flag = False
            else:
                plt.plot([nodes[j].xm], [nodes[j].ym], 'kd')
        '''
        plt.legend()
        plt.xlabel('X/m')
        plt.ylabel('Y/m')
        plt.show()

class Leach(object):
    """ Leach """
    # Optimal selection probablitity of a node to become cluster head
    p       = 0.1
    period  = int(1/p)
    heads   = None
    members = None
    cluster = None
    r       = 0
    rmax    = 500  #9999 # default maximum round+++++++++++++++++
    r_empty = 0

    def show_cluster():
        fig = plt.figure()
        plt.xlabel('X/m')
        plt.ylabel('Y/m')
        icon = ['o', '*', '.', 'x', '+', 's']
        color = ['r', 'b', 'g', 'c', 'y', 'm']
        i = 0
        nodes = WSN.nodes
        for key, value in Leach.cluster.items():
            cluster_head = nodes[int(key)]
            for index in value:
                plt.plot([cluster_head.xm, nodes[index].xm], [cluster_head.ym, nodes[index].ym], c=color[i % 6], marker=icon[i % 5], alpha=0.4)
                if index >= WSN.n:
                    #plt.plot([nodes[index].xm], [nodes[index].ym], 'dk')
                    i += 1
        plt.show()

    def optimum_number_of_clusters():
        """optimization """
        N = WSN.n - WSN.n_dead
        M = np.sqrt(WSN.xm * WSN.ym)
        d_toBS = np.sqrt((WSN.sink.xm - WSN.xm) ** 2 + (WSN.sink.ym - WSN.ym) ** 2)
        k_opt = (np.sqrt(N) / np.sqrt(2 * np.pi) * np.sqrt(WSN.Efs / WSN.Emp) * M / (d_toBS ** 2))
        p = int(k_opt) / N
        return p

    def cluster_head_selection():   #
        """ CH_selection """
        nodes   = WSN.nodes
        n       = WSN.n
        heads   = Leach.heads = []
        members = Leach.members = []
        p       = Leach.p
        r       = Leach.r
        period  = Leach.period
        Tn      = p / (1 - p * (r % period))
        print("The number of iterations:", Leach.r)
        for i in range(n):
            # After the energy dissipated in a given node reached a set threshold,
            # that node was considered dead for the remainder of the simulation.
            if nodes[i].energy > Node.energy_threshold: #
                if nodes[i].G == 0: #
                    temp_rand = np.random.random()
#                    print(temp_rand)
                    if temp_rand <= Tn:
                        nodes[i].type = "CH"
                        nodes[i].G = 1
                        heads.append(nodes[i])
                        # Announce cluster-head status, wait for join-request messages
                        max_dis = np.sqrt(WSN.xm ** 2 +  WSN.ym ** 2)
                        nodes[i].energy -= WSN.trans_energy(WSN.CM, max_dis)
                        WSN.d_data+=1;
                        print(np.int64.max(nodes[i].energy)*100);
                if nodes[i].type == "N":
                    members.append(nodes[i])
        #m_n = WSN.m_n
        '''
        for i in range(m_n):
            j = n + i
            members.append(nodes[j])
        '''
        if not heads:
            Leach.r_empty += 1
            print("---> No cluster head was found this round!")
            # Leach.cluster_head_selection()
        print("The number of CHs is:", len(heads), "The number of Alive Node:", (WSN.n - WSN.n_dead))
        print("The number of Delivery Packets:", (WSN.packets_data_d/WSN.d_data))
        print("Rate of Delay:", ((WSN.packets_data_d)/(WSN.d_data*WSN.n)*WSN.n))
        print("Rate of Throughput (kbps):", (WSN.packets_data_d/Leach.rmax))

        return None # heads, members
    def cluster_formation():
        """ cluster_formation """
        nodes   = WSN.nodes
        heads   = Leach.heads
        members = Leach.members
        cluster = Leach.cluster = {}
        if not heads:
            return None
        for head in heads:
            cluster[str(head.id)] = []
        # print("clustering:", cluster)
        for member in members:
            min_dis = np.sqrt(WSN.xm ** 2 +  WSN.ym ** 2)
            head_id = None
            # wait for cluster-head announcements
            member.energy -= WSN.ERX * WSN.CM * len(heads)
            for head in heads:
                tmp = WSN.dist(member, head)
                if tmp <= min_dis:
                    min_dis = tmp
                    head_id = head.id
            member.head_id = head_id
            # send join-request messages to chosen cluster-head
            member.energy -= WSN.trans_energy(WSN.CM, min_dis)
            # wait for join-request messages
            head = nodes[head_id]
            head.energy -= WSN.ERX * WSN.CM
            cluster[str(head_id)].append(member.id)
        # Create a TDMA schedule and this schedule is broadcast back to the nodes in the cluster.
        for key, values in cluster.items():
            head = nodes[int(key)]
            if not values:
                # If there are cluster members, the CH sends schedule by broadcasting
                max_dis = np.sqrt(WSN.xm ** 2 +  WSN.ym ** 2)
                head.energy -= WSN.trans_energy(WSN.CM, max_dis)
                WSN.d_data+=1;
                for x in values:
                    member = nodes[int(x)]
                    # wait for schedule from cluster-head
                    member.energy -= WSN.ERX * WSN.CM
                    #WSN.packets_data_s+=1;
                    #WSN.d_data+=1;
#        print(cluster)
        return None # cluster

    def set_up_phase():
        Leach.cluster_head_selection()
        Leach.cluster_formation()

    def steady_state_phase():
        """ steady_state_phase """
        nodes   = WSN.nodes
        cluster = Leach.cluster
        if not cluster:
            return None
        for key, values in cluster.items():
            head     = nodes[int(key)]
            n_member = len(values)
            for x in values:
                member = nodes[int(x)]
                dis    = WSN.dist(member, head)
                member.energy -= WSN.trans_energy(WSN.DM, dis)
                head.energy   -= WSN.ERX * WSN.DM
            d_h2s = WSN.dist(head, WSN.sink) # The distance of from head to sink
            if n_member == 0:
                energy = WSN.trans_energy(WSN.DM, d_h2s)
            else:
                new_data = WSN.DM * (n_member + 1)
                E_DA     = WSN.EDA * new_data
                if WSN.f_r == 0:
                    new_data_ = WSN.DM
                else:
                    new_data_ = new_data * WSN.f_r
                E_Trans  = WSN.trans_energy(new_data_, d_h2s)
                energy = E_DA + E_Trans
            head.energy -= energy

    def leach():
        Leach.set_up_phase()
        Leach.steady_state_phase()

    def run_leach():
        for r in range(Leach.rmax):
            Leach.r = r
            nodes   = WSN.nodes
            if (r % Leach.period) == 0:
                print("==============================")
                for node in nodes:
                    node.G = 0
            for node in nodes:
                node.type = "N"
            Leach.leach()
            WSN.node_state(r)
            if WSN.flag_all_dead:
                print("==============================")
                break
        Leach.show_cluster()
def main():
    Node.init_nodes()
    #Node.init_malicious_nodes()
    Node.plot_wsn()
    Leach.run_leach()
    # print("The first node died in Round %d!" % (WSN.round_first_dead))
    # print("The network stop working in Round %d!" % (WSN.round_net_stop))
    # print("All nodes died in Round %d!" % (WSN.round_all_dead))

if __name__ == '__main__':
    main()